@@ -71,8 +71,8 @@ def get_image_schema():
                   'passed to the client via stdin.'))
 @utils.arg('--progress', action='store_true', default=False,
            help=_('Show upload progress bar.'))
-@utils.arg('--backend', metavar='<STORE>',
-           default=utils.env('OS_IMAGE_BACKEND', default=None),
+@utils.arg('--store', metavar='<STORE>',
+           default=utils.env('OS_IMAGE_STORE', default=None),
            help='Backend store to upload image to.')
 @utils.on_data_require_fields(DATA_FIELDS)
 def do_image_create(gc, args):
@@ -90,12 +90,12 @@ def do_image_create(gc, args):
         key, value = datum.split('=', 1)
         fields[key] = value
 
-    backend = args.backend
+    backend = args.store
 
     file_name = fields.pop('file', None)
     using_stdin = not sys.stdin.isatty()
-    if args.backend and not (file_name or using_stdin):
-        utils.exit("--backend option should only be provided with --file "
+    if args.store and not (file_name or using_stdin):
+        utils.exit("--store option should only be provided with --file "
                    "option or stdin.")
 
     if backend:
@@ -108,7 +108,7 @@ def do_image_create(gc, args):
     image = gc.images.create(**fields)
     try:
         if utils.get_data_file(args) is not None:
-            backend = fields.get('backend', None)
+            backend = fields.get('store', None)
             args.id = image['id']
             args.size = None
             do_image_upload(gc, args)
@@ -147,9 +147,31 @@ def do_image_create(gc, args):
                   'record if no import-method and no data is supplied'))
 @utils.arg('--uri', metavar='<IMAGE_URL>', default=None,
            help=_('URI to download the external image.'))
-@utils.arg('--backend', metavar='<STORE>',
-           default=utils.env('OS_IMAGE_BACKEND', default=None),
+@utils.arg('--store', metavar='<STORE>',
+           default=utils.env('OS_IMAGE_STORE', default=None),
            help='Backend store to upload image to.')
+@utils.arg('--stores', metavar='<STORES>',
+           default=utils.env('OS_IMAGE_STORES', default=None),
+           help=_('Stores to upload image to if multi-stores import '
+                  'available. Comma separated list. Available stores can be '
+                  'listed with "stores-info" call.'))
+@utils.arg('--all-stores', type=strutils.bool_from_string,
+           metavar='[True|False]',
+           default=None,
+           dest='os_all_stores',
+           help=_('"all-stores" can be ued instead of "stores"-list to '
+                  'indicate that image should be imported into all available '
+                  'stores.'))
+@utils.arg('--allow-failure', type=strutils.bool_from_string,
+           metavar='[True|False]',
+           dest='os_allow_failure',
+           default=utils.env('OS_IMAGE_ALLOW_FAILURE', default=True),
+           help=_('Indicator if all stores listed (or available) must '
+                  'succeed. "True" by default meaning that we allow some '
+                  'stores to fail and the status can be monitored from the '
+                  'image metadata. If this is set to "False" the import will '
+                  'be reverted should any of the uploads fail. Only usable '
+                  'with "stores" or "all-stores".'))
 @utils.on_data_require_fields(DATA_FIELDS)
 def do_image_create_via_import(gc, args):
     """EXPERIMENTAL: Create a new image via image import.
@@ -188,6 +210,10 @@ def do_image_create_via_import(gc, args):
     if args.import_method is None and (file_name or using_stdin):
         args.import_method = 'glance-direct'
 
+    if args.import_method == 'copy-image':
+        utils.exit("Import method 'copy-image' cannot be used "
+                   "while creating the image.")
+
     # determine whether the requested import method is valid
     import_methods = gc.images.get_import_info().get('import-methods')
     if args.import_method and args.import_method not in import_methods.get(
@@ -198,9 +224,24 @@ def do_image_create_via_import(gc, args):
 
     # determine if backend is valid
     backend = None
-    if args.backend:
-        backend = args.backend
+    stores = getattr(args, "stores", None)
+    all_stores = getattr(args, "os_all_stores", None)
+
+    if (args.store and (stores or all_stores)) or (stores and all_stores):
+        utils.exit("Only one of --store, --stores and --all-stores can be "
+                   "provided")
+    elif args.store:
+        backend = args.store
+        # determine if backend is valid
         _validate_backend(backend, gc)
+    elif stores:
+        # NOTE(jokke): Making sure here that we do not include the stores in
+        # the create call
+        fields.pop("stores")
+        stores = str(stores).split(',')
+        for store in stores:
+            # determine if backend is valid
+            _validate_backend(store, gc)
 
     # make sure we have all and only correct inputs for the requested method
     if args.import_method is None:
@@ -209,8 +250,16 @@ def do_image_create_via_import(gc, args):
                        "method.")
     if args.import_method == 'glance-direct':
         if backend and not (file_name or using_stdin):
-            utils.exit("--backend option should only be provided with --file "
+            utils.exit("--store option should only be provided with --file "
                        "option or stdin for the glance-direct import method.")
+        if stores and not (file_name or using_stdin):
+            utils.exit("--stores option should only be provided with --file "
+                       "option or stdin for the glance-direct import method.")
+        if all_stores and not (file_name or using_stdin):
+            utils.exit("--all-stores option should only be provided with "
+                       "--file option or stdin for the glance-direct import "
+                       "method.")
+
         if args.uri:
             utils.exit("You cannot specify a --uri with the glance-direct "
                        "import method.")
@@ -225,8 +274,14 @@ def do_image_create_via_import(gc, args):
                        "for the glance-direct import method.")
     if args.import_method == 'web-download':
         if backend and not args.uri:
-            utils.exit("--backend option should only be provided with --uri "
+            utils.exit("--store option should only be provided with --uri "
                        "option for the web-download import method.")
+        if stores and not args.uri:
+            utils.exit("--stores option should only be provided with --uri "
+                       "option for the web-download import method.")
+        if all_stores and not args.uri:
+            utils.exit("--all-stores option should only be provided with "
+                       "--uri option for the web-download import method.")
         if not args.uri:
             utils.exit("URI is required for web-download import method. "
                        "Please use '--uri <uri>'.")
@@ -246,6 +301,7 @@ def do_image_create_via_import(gc, args):
                 args.size = None
                 do_image_stage(gc, args)
             args.from_create = True
+            args.stores = stores
             do_image_import(gc, args)
         image = gc.images.get(args.id)
     finally:
@@ -267,7 +323,7 @@ def _validate_backend(backend, gc):
                 break
 
         if not valid_backend:
-            utils.exit("Backend '%s' is not valid for this cloud. Valid "
+            utils.exit("Store '%s' is not valid for this cloud. Valid "
                        "values can be retrieved with stores-info command." %
                        backend)
 
@@ -327,6 +383,9 @@ def do_image_update(gc, args):
            action='append', dest='properties', default=[])
 @utils.arg('--checksum', metavar='<CHECKSUM>',
            help=_('Displays images that match the MD5 checksum.'))
+@utils.arg('--hash', dest='os_hash_value', default=None,
+           metavar='<HASH_VALUE>',
+           help=_('Displays images that match the specified hash value.'))
 @utils.arg('--tag', metavar='<TAG>', action='append',
            help=_("Filter images by a user-defined tag."))
 @utils.arg('--sort-key', default=[], action='append',
@@ -358,7 +417,7 @@ def do_image_update(gc, args):
 def do_image_list(gc, args):
     """List images you can access."""
     filter_keys = ['visibility', 'member_status', 'owner', 'checksum', 'tag',
-                   'os_hidden']
+                   'os_hidden', 'os_hash_value']
     filter_items = [(key, getattr(args, key)) for key in filter_keys]
 
     if args.properties:
@@ -500,6 +559,24 @@ def do_stores_info(gc, args):
         utils.print_dict(stores_info)
 
 
+@utils.arg('id', metavar='<IMAGE_ID>', help=_('ID of image to update.'))
+@utils.arg('--store', metavar='<STORE_ID>', required=True,
+           help=_('Store to delete image from.'))
+def do_stores_delete(gc, args):
+    """Delete image from specific store."""
+    try:
+        gc.images.delete_from_store(args.store, args.id)
+    except exc.HTTPNotFound:
+        utils.exit('Multi Backend support is not enabled or Image/store not '
+                   'found.')
+    except (exc.HTTPForbidden, exc.HTTPException) as e:
+        msg = ("Unable to delete image '%s' from store '%s'. (%s)" % (
+               args.id,
+               args.store,
+               e))
+        utils.exit(msg)
+
+
 @utils.arg('--allow-md5-fallback', action='store_true',
            default=utils.env('OS_IMAGE_ALLOW_MD5_FALLBACK', default=False),
            help=_('If os_hash_algo and os_hash_value properties are available '
@@ -508,7 +585,7 @@ def do_stores_info(gc, args):
                   'available on the client, the download will fail. Use this '
                   'flag to indicate that in such a case the legacy MD5 image '
                   'checksum should be used to validate the downloaded data. '
-                  'You can also set the enviroment variable '
+                  'You can also set the environment variable '
                   'OS_IMAGE_ALLOW_MD5_FALLBACK to any value to activate this '
                   'option.'))
 @utils.arg('--file', metavar='<FILE>',
@@ -556,14 +633,14 @@ def do_image_download(gc, args):
            help=_('Show upload progress bar.'))
 @utils.arg('id', metavar='<IMAGE_ID>',
            help=_('ID of image to upload data to.'))
-@utils.arg('--backend', metavar='<STORE>',
-           default=utils.env('OS_IMAGE_BACKEND', default=None),
+@utils.arg('--store', metavar='<STORE>',
+           default=utils.env('OS_IMAGE_STORE', default=None),
            help='Backend store to upload image to.')
 def do_image_upload(gc, args):
     """Upload data for a specific image."""
     backend = None
-    if args.backend:
-        backend = args.backend
+    if args.store:
+        backend = args.store
         # determine if backend is valid
         _validate_backend(backend, gc)
 
@@ -611,22 +688,59 @@ def do_image_stage(gc, args):
            help=_('URI to download the external image.'))
 @utils.arg('id', metavar='<IMAGE_ID>',
            help=_('ID of image to import.'))
-@utils.arg('--backend', metavar='<STORE>',
-           default=utils.env('OS_IMAGE_BACKEND', default=None),
+@utils.arg('--store', metavar='<STORE>',
+           default=utils.env('OS_IMAGE_STORE', default=None),
            help='Backend store to upload image to.')
+@utils.arg('--stores', metavar='<STORES>',
+           default=utils.env('OS_IMAGE_STORES', default=None),
+           help='Stores to upload image to if multi-stores import available.')
+@utils.arg('--all-stores', type=strutils.bool_from_string,
+           metavar='[True|False]',
+           default=None,
+           dest='os_all_stores',
+           help=_('"all-stores" can be ued instead of "stores"-list to '
+                  'indicate that image should be imported all available '
+                  'stores.'))
+@utils.arg('--allow-failure', type=strutils.bool_from_string,
+           metavar='[True|False]',
+           dest='os_allow_failure',
+           default=utils.env('OS_IMAGE_ALLOW_FAILURE', default=True),
+           help=_('Indicator if all stores listed (or available) must '
+                  'succeed. "True" by default meaning that we allow some '
+                  'stores to fail and the status can be monitored from the '
+                  'image metadata. If this is set to "False" the import will '
+                  'be reverted should any of the uploads fail. Only usable '
+                  'with "stores" or "all-stores".'))
 def do_image_import(gc, args):
     """Initiate the image import taskflow."""
-    backend = None
-    if args.backend:
-        backend = args.backend
+    backend = getattr(args, "store", None)
+    stores = getattr(args, "stores", None)
+    all_stores = getattr(args, "os_all_stores", None)
+    allow_failure = getattr(args, "os_allow_failure", True)
+
+    if not getattr(args, 'from_create', False):
+        if (args.store and (stores or all_stores)) or (stores and all_stores):
+            utils.exit("Only one of --store, --stores and --all-stores can be "
+                       "provided")
+        elif args.store:
+            backend = args.store
+            # determine if backend is valid
+            _validate_backend(backend, gc)
+        elif stores:
+            stores = str(stores).split(',')
+
         # determine if backend is valid
-        _validate_backend(backend, gc)
+        if stores:
+            for store in stores:
+                _validate_backend(store, gc)
 
     if getattr(args, 'from_create', False):
         # this command is being called "internally" so we can skip
         # validation -- just do the import and get out of here
         gc.images.image_import(args.id, args.import_method, args.uri,
-                               backend=backend)
+                               backend=backend,
+                               stores=stores, all_stores=all_stores,
+                               allow_failure=allow_failure)
         return
 
     # do input validation
@@ -646,6 +760,10 @@ def do_image_import(gc, args):
         utils.exit("Import method should be 'web-download' if URI is "
                    "provided.")
 
+    if args.import_method == 'copy-image' and not (stores or all_stores):
+        utils.exit("Provide either --stores or --all-stores for "
+                   "'copy-image' import method.")
+
     # check image properties
     image = gc.images.get(args.id)
     container_format = image.get('container_format')
@@ -663,10 +781,16 @@ def do_image_import(gc, args):
         if image_status != 'queued':
             utils.exit("The 'web-download' import method can only be applied "
                        "to an image in status 'queued'")
+    if args.import_method == 'copy-image':
+        if image_status != 'active':
+            utils.exit("The 'copy-image' import method can only be used on "
+                       "an image with status 'active'.")
 
     # finally, do the import
     gc.images.image_import(args.id, args.import_method, args.uri,
-                           backend=backend)
+                           backend=backend,
+                           stores=stores, all_stores=all_stores,
+                           allow_failure=allow_failure)
 
     image = gc.images.get(args.id)
     utils.print_image(image)
@@ -719,15 +843,15 @@ def do_image_reactivate(gc, args):
 @utils.arg('tag_value', metavar='<TAG_VALUE>',
            help=_('Value of the tag.'))
 def do_image_tag_update(gc, args):
-        """Update an image with the given tag."""
-        if not (args.image_id and args.tag_value):
-            utils.exit('Unable to update tag. Specify image_id and tag_value')
-        else:
-            gc.image_tags.update(args.image_id, args.tag_value)
-            image = gc.images.get(args.image_id)
-            image = [image]
-            columns = ['ID', 'Tags']
-            utils.print_list(image, columns)
+    """Update an image with the given tag."""
+    if not (args.image_id and args.tag_value):
+        utils.exit('Unable to update tag. Specify image_id and tag_value')
+    else:
+        gc.image_tags.update(args.image_id, args.tag_value)
+        image = gc.images.get(args.image_id)
+        image = [image]
+        columns = ['ID', 'Tags']
+        utils.print_list(image, columns)
 
 
 @utils.arg('image_id', metavar='<IMAGE_ID>',
@@ -747,16 +871,30 @@ def do_image_tag_delete(gc, args):
 @utils.arg('--metadata', metavar='<STRING>', default='{}',
            help=_('Metadata associated with the location. '
                   'Must be a valid JSON object (default: %(default)s)'))
+@utils.arg('--checksum', metavar='<STRING>',
+           help=_('md5 checksum of image contents'))
+@utils.arg('--hash-algo', metavar='<STRING>',
+           help=_('Multihash algorithm'))
+@utils.arg('--hash-value', metavar='<STRING>',
+           help=_('Multihash value'))
 @utils.arg('id', metavar='<IMAGE_ID>',
            help=_('ID of image to which the location is to be added.'))
 def do_location_add(gc, args):
     """Add a location (and related metadata) to an image."""
+    validation_data = {}
+    if args.checksum:
+        validation_data['checksum'] = args.checksum
+    if args.hash_algo:
+        validation_data['os_hash_algo'] = args.hash_algo
+    if args.hash_value:
+        validation_data['os_hash_value'] = args.hash_value
     try:
         metadata = json.loads(args.metadata)
     except ValueError:
         utils.exit('Metadata is not a valid JSON object.')
     else:
-        image = gc.images.add_location(args.id, args.url, metadata)
+        image = gc.images.add_location(args.id, args.url, metadata,
+                                       validation_data=validation_data)
         utils.print_dict(image)
 
 
